'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var reduxActions = require('redux-actions');

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _objectSpread(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};
    var ownKeys = Object.keys(source);

    if (typeof Object.getOwnPropertySymbols === 'function') {
      ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) {
        return Object.getOwnPropertyDescriptor(source, sym).enumerable;
      }));
    }

    ownKeys.forEach(function (key) {
      _defineProperty(target, key, source[key]);
    });
  }

  return target;
}

var isObject = function isObject(o) {
  return typeof o === 'object' && o !== null;
};

var reduceReducers = function reduceReducers() {
  for (var _len = arguments.length, reducers = new Array(_len), _key = 0; _key < _len; _key++) {
    reducers[_key] = arguments[_key];
  }

  var initialState = null;
  return function (prevState, value) {
    for (var _len2 = arguments.length, args = new Array(_len2 > 2 ? _len2 - 2 : 0), _key2 = 2; _key2 < _len2; _key2++) {
      args[_key2 - 2] = arguments[_key2];
    }

    var prevStateIsUndefined = typeof prevState === 'undefined';
    var valueIsUndefined = typeof value === 'undefined';

    if (prevStateIsUndefined && valueIsUndefined && initialState) {
      return initialState;
    }

    var x = prevStateIsUndefined && !valueIsUndefined && initialState;
    return reducers.reduce(function (newState, reducer, index) {
      if (typeof reducer === 'undefined') {
        throw new TypeError("An undefined reducer was passed in at index " + index);
      }

      return reducer.apply(void 0, [newState, value].concat(args));
    }, x ? initialState : prevState);
  };
};

var FETCHABLE_STATUS = {
  INITIAL: 'INITIAL',
  LOADING: 'LOADING',
  SUCCESS: 'SUCCESS',
  FAILURE: 'FAILURE'
};
var validateInject = function validateInject(injected) {
  if (!Array.isArray(injected)) {
    throw Error('Inject should be an array of dependency names');
  }

  var hasNonStringInjects = injected.map(function (dep) {
    return typeof dep !== 'string';
  }).filter(Boolean).length > 0;

  if (hasNonStringInjects) {
    throw Error('Injected dependency names should be strings');
  }
};
var validateModel = function validateModel(_ref) {
  var name = _ref.name,
      state = _ref.state,
      inject = _ref.inject,
      actions = _ref.actions,
      reactions = _ref.reactions;
  if (!name) throw Error('Model should have a name');

  if (state && !actions && !reactions) {
    throw Error('Model with state should have reducers');
  }

  if (inject) validateInject(inject);
};
var handleThunks = function handleThunks(thunks, dependencies) {
  return Object.entries(thunks).reduce(function (acc, _ref2) {
    var actionName = _ref2[0],
        thunk = _ref2[1];

    acc[actionName] = function () {
      for (var _len3 = arguments.length, args = new Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
        args[_key3] = arguments[_key3];
      }

      return thunk.apply(void 0, args.concat([dependencies]));
    };

    return acc;
  }, {});
};

var createFetchableReducers = function createFetchableReducers(_ref3) {
  var _defaultReducers;

  var types = _ref3.types,
      successField = _ref3.successField,
      overrides = _ref3.overrides;

  if (!successField) {
    throw Error('You must provide the name of the field that is used for success payload');
  }

  var defaultReducers = (_defaultReducers = {}, _defaultReducers[types.loading] = function (state) {
    var _objectSpread2;

    return _objectSpread({}, state, (_objectSpread2 = {}, _objectSpread2[successField] = _objectSpread({}, state[successField], {
      status: FETCHABLE_STATUS.LOADING,
      error: null
    }), _objectSpread2));
  }, _defaultReducers[types.success] = function (state, action) {
    var _objectSpread3;

    return _objectSpread({}, state, (_objectSpread3 = {}, _objectSpread3[successField] = {
      data: action.payload,
      status: FETCHABLE_STATUS.SUCCESS,
      error: null
    }, _objectSpread3));
  }, _defaultReducers[types.failure] = function (state, action) {
    var _objectSpread4;

    return _objectSpread({}, state, (_objectSpread4 = {}, _objectSpread4[successField] = _objectSpread({}, state[successField], {
      status: FETCHABLE_STATUS.FAILURE,
      error: action.payload
    }), _objectSpread4));
  }, _defaultReducers);
  if (!isObject(overrides)) return defaultReducers; // Apply possible overrides

  return Object.entries(types).reduce(function (acc, _ref4) {
    var k = _ref4[0],
        v = _ref4[1];
    acc[v] = overrides[k] ? reduceReducers(defaultReducers[v], overrides[k]) : defaultReducers[v];
    return acc;
  }, {});
};

var createFetchableAction = function createFetchableAction(types) {
  var action = reduxActions.createAction(types.loading);
  action.success = reduxActions.createAction(types.success);
  action.fail = reduxActions.createAction(types.failure);
  action.init = reduxActions.createAction(types.init);
  return action;
};

function handleFetchableAction(args, actionName, modelName) {
  var _types;

  var typePrefix = modelName + "/" + actionName;
  var t = {
    loading: typePrefix,
    success: typePrefix + "/success",
    failure: typePrefix + "/failure",
    // NOTE: Also provide `init` for cases where loading should not be
    // dispatched as the default action
    init: typePrefix + "/init"
  };
  var action = createFetchableAction(t); // User can either provide only reducer field name for success case
  // or reducer overrides for `loading` / `success` / `failure` cases

  var reducers = createFetchableReducers({
    types: t,
    successField: args.length > 0 ? args[0] : null,
    overrides: args.length > 1 ? args[1] : {}
  }); // Return types that are inlined to the other types instead of accessing them
  // via `types.fetchSomething.success` you access them normally
  // without object notation `types.fetchSomethingSuccess`
  // (where `fetchSomething` is the action name)

  var types = (_types = {}, _types[actionName] = typePrefix, _types[actionName + "Init"] = typePrefix + "/init", _types[actionName + "Success"] = typePrefix + "/success", _types[actionName + "Failure"] = typePrefix + "/failure", _types);
  return {
    action: action,
    reducers: reducers,
    types: types
  };
}
var FETCHABLE_ACTION_IDENTIFIER = '__IS_FETCHABLE_ACTION__';
var isFetchableAction = function isFetchableAction(x) {
  return x && Object.prototype.hasOwnProperty.call(x, FETCHABLE_ACTION_IDENTIFIER);
};

var createModel = function createModel(model) {
  validateModel(model);

  var initialState = _objectSpread({}, model.state);

  var dependencies = {};
  var types = {};
  var reducerHandlers = {};
  var actions = {};
  var reducer;
  var sagas = []; // Mark injected deps to be filled later

  if (model.inject) {
    model.inject.forEach(function (depName) {
      dependencies[depName] = null;
    });
  } // Create types, actions and reducer handlers from `model.actions`
  // We need to provide `initialState` since it might be used in reducers


  Object.entries(model.actions({
    initialState: initialState
  })).forEach(function (_ref) {
    var actionName = _ref[0],
        reducerHandler = _ref[1];

    if (isFetchableAction(reducerHandler)) {
      // Handle async API actions
      var fetchableX = handleFetchableAction(reducerHandler.args, actionName, model.name); // Inline fetchable types

      Object.entries(fetchableX.types).forEach(function (_ref2) {
        var k = _ref2[0],
            v = _ref2[1];
        types[k] = v;
      });
      actions[actionName] = fetchableX.action;
      reducerHandlers = _objectSpread({}, reducerHandlers, fetchableX.reducers);
    } else {
      // Register action type
      var actionType = model.name + "/" + actionName;
      types[actionName] = actionType; // Create basic action

      actions[actionName] = reduxActions.createAction(actionType);
      if (reducerHandler) reducerHandlers[actionType] = reducerHandler;
    }
  }, {}); // Handle thunks

  if (model.thunks) {
    actions = _objectSpread({}, actions, handleThunks(model.thunks, dependencies));
  }

  var selectors = {}; // Add selectors defined by user

  if (model.selectors) {
    selectors = model.selectors({
      name: model.name
    });
  } // Add simple `get` for selecting state fields by name


  var getSelector = function getSelector(key) {
    return function (state) {
      var statePart = state[model.name];

      if (!Object.prototype.hasOwnProperty.call(statePart, key)) {
        throw Error("You tried to select a non-existent field '" + key + "' from state");
      }

      return statePart[key];
    };
  };

  selectors = _objectSpread({}, selectors, {
    get: getSelector
  }); // Fills in the dependencies that were requested when calling inject.

  var _fillDeps = function _fillDeps(deps) {
    Object.keys(dependencies).forEach(function (dep) {
      if (deps[dep]) {
        dependencies[dep] = {
          // Don't spread all of dep but only what's needed
          types: deps[dep].types,
          actions: deps[dep].actions,
          selectors: deps[dep].selectors
        };
      } else {
        throw Error("There is no dependendy called '" + dep + "' for " + model.name + " model." // eslint-disable-line
        );
      }
    });
  }; // Run reducer and sagas functions with the necessary data.


  var _run = function _run() {
    // Run model reactions function to get rest of the reducer
    if (model.reactions) {
      var fromDeps = model.reactions({
        initialState: initialState,
        deps: dependencies
      });
      reducerHandlers = _objectSpread({}, reducerHandlers, fromDeps);
    }

    reducer = reduxActions.handleActions(reducerHandlers, initialState);

    if (model.sagas) {
      sagas = model.sagas({
        types: types,
        deps: dependencies
      });
    }
  }; // Get reducer for the model.


  var getReducer = function getReducer() {
    return reducer;
  }; // Get sagas for the model.


  var getSagas = function getSagas() {
    return sagas;
  };

  return {
    name: model.name,
    initialState: model.state,
    types: types,
    actions: actions,
    selectors: selectors,
    getSagas: getSagas,
    getReducer: getReducer,
    _run: _run,
    _fillDeps: _fillDeps
  };
}; // Create the final models with dependencies injected

var initModels = function initModels(models) {
  if (models === void 0) {
    models = [];
  }

  var modelsByName = models.reduce(function (acc, val) {
    acc[val.name] = val;
    return acc;
  }, {}); // @ts-ignore

  Object.values(modelsByName).forEach(function (model) {
    model._fillDeps(modelsByName);

    model._run(); // Users don't need these props so delete them


    delete model._fillDeps;
    delete model._run;
  });

  var getAllReducers = function getAllReducers() {
    return Object.values(modelsByName).reduce(function (acc, val) {
      acc[val.name] = val.getReducer();
      return acc;
    }, {});
  };

  var getAllSagas = function getAllSagas() {
    return Object.values(modelsByName).reduce(function (acc, val) {
      return acc.concat.apply(acc, val.getSagas());
    }, []);
  };

  return _objectSpread({}, modelsByName, {
    allReducers: getAllReducers(),
    allSagas: getAllSagas()
  });
};
var FetchableStatus = FETCHABLE_STATUS;
var fetchable = {
  value: function value(initialValue) {
    return {
      data: initialValue,
      status: FETCHABLE_STATUS.INITIAL,
      error: null
    };
  },
  action: function action() {
    var _ref3;

    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    return _ref3 = {}, _ref3[FETCHABLE_ACTION_IDENTIFIER] = true, _ref3.args = args, _ref3;
  }
};

exports.createModel = createModel;
exports.initModels = initModels;
exports.FetchableStatus = FetchableStatus;
exports.fetchable = fetchable;
